# githelp.py - Try to map Git commands to Mercurial equivalents.
#
# Copyright 2013 Facebook, Inc.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
"""try mapping git commands to Mercurial commands

Tries to map a given git command to a Mercurial command:

  $ hg githelp -- git checkout master
  hg update master

If an unknown command or parameter combination is detected, an error is
produced.
"""

from __future__ import absolute_import

import getopt
import re

from mercurial.i18n import _
from mercurial import (
    error,
    fancyopts,
    registrar,
    util,
)

# Note for extension authors: ONLY specify testedwith = 'ships-with-hg-core' for
# extensions which SHIP WITH MERCURIAL. Non-mainline extensions should
# be specifying the version(s) of Mercurial they are tested with, or
# leave the attribute unspecified.
testedwith = 'ships-with-hg-core'

cmdtable = {}
command = registrar.command(cmdtable)

def convert(s):
    if s.startswith("origin/"):
        return s[7:]
    if 'HEAD' in s:
        s = s.replace('HEAD', '.')
    # HEAD~ in git is .~1 in mercurial
    s = re.sub('~$', '~1', s)
    return s

@command('^githelp|git', [
    ], _('hg githelp'))
def githelp(ui, repo, *args, **kwargs):
    '''suggests the Mercurial equivalent of the given git command

    Usage: hg githelp -- <git command>
    '''

    if len(args) == 0 or (len(args) == 1 and args[0] =='git'):
        raise error.Abort(_('missing git command - '
                            'usage: hg githelp -- <git command>'))

    if args[0] == 'git':
        args = args[1:]

    cmd = args[0]
    if not cmd in gitcommands:
        raise error.Abort("error: unknown git command %s" % (cmd))

    ui.pager('githelp')
    args = args[1:]
    return gitcommands[cmd](ui, repo, *args, **kwargs)

def parseoptions(ui, cmdoptions, args):
    cmdoptions = list(cmdoptions)
    opts = {}
    args = list(args)
    while True:
        try:
            args = fancyopts.fancyopts(list(args), cmdoptions, opts, True)
            break
        except getopt.GetoptError as ex:
            flag = None
            if "requires argument" in ex.msg:
                raise
            if ('--' + ex.opt) in ex.msg:
                flag = '--' + ex.opt
            elif ('-' + ex.opt) in ex.msg:
                flag = '-' + ex.opt
            else:
                raise error.Abort("unknown option %s" % ex.opt)
            try:
                args.remove(flag)
            except Exception:
                raise error.Abort(
                    "unknown option {0} packed with other options\n"
                    "Please try passing the option as it's own flag: -{0}" \
                    .format(ex.opt))

            ui.warn(_("ignoring unknown option %s\n") % flag)

    args = list([convert(x) for x in args])
    opts = dict([(k, convert(v)) if isinstance(v, str) else (k, v)
                                 for k, v in opts.iteritems()])

    return args, opts

class Command(object):
    def __init__(self, name):
        self.name = name
        self.args = []
        self.opts = {}

    def __str__(self):
        cmd = "hg " + self.name
        if self.opts:
            for k, values in sorted(self.opts.iteritems()):
                for v in values:
                    if v:
                        cmd += " %s %s" % (k, v)
                    else:
                        cmd += " %s" % (k,)
        if self.args:
            cmd += " "
            cmd += " ".join(self.args)
        return cmd

    def append(self, value):
        self.args.append(value)

    def extend(self, values):
        self.args.extend(values)

    def __setitem__(self, key, value):
        values = self.opts.setdefault(key, [])
        values.append(value)

    def __and__(self, other):
        return AndCommand(self, other)

class AndCommand(object):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __str__(self):
        return "%s && %s" % (self.left, self.right)

    def __and__(self, other):
        return AndCommand(self, other)

def add(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('A', 'all', None, ''),
        ('p', 'patch', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    if (opts.get('patch')):
        ui.status(_("note: Mercurial will commit when complete, "
                    "as there is no staging area in Mercurial\n\n"))
        cmd = Command('commit --interactive')
    else:
        cmd = Command("add")

        if not opts.get('all'):
            cmd.extend(args)
        else:
            ui.status(_("note: use hg addremove to remove files that have "
                        "been deleted.\n\n"))

    ui.status((str(cmd)), "\n")

def am(ui, repo, *args, **kwargs):
    cmdoptions=[
    ]
    args, opts = parseoptions(ui, cmdoptions, args)
    cmd = Command('import')
    ui.status(str(cmd), "\n")

def apply(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('p', 'p', int, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('import --no-commit')
    if (opts.get('p')):
        cmd['-p'] = opts.get('p')
    cmd.extend(args)

    ui.status((str(cmd)), "\n")

def bisect(ui, repo, *args, **kwargs):
    ui.status(_("See 'hg help bisect' for how to use bisect.\n\n"))

def blame(ui, repo, *args, **kwargs):
    cmdoptions = [
    ]
    args, opts = parseoptions(ui, cmdoptions, args)
    cmd = Command('annotate -udl')
    cmd.extend([convert(v) for v in args])
    ui.status((str(cmd)), "\n")

def branch(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'set-upstream', None, ''),
        ('', 'set-upstream-to', '', ''),
        ('d', 'delete', None, ''),
        ('D', 'delete', None, ''),
        ('m', 'move', None, ''),
        ('M', 'move', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command("bookmark")

    if opts.get('set_upstream') or opts.get('set_upstream_to'):
        ui.status(_("Mercurial has no concept of upstream branches\n"))
        return
    elif opts.get('delete'):
        cmd = Command("strip")
        for branch in args:
            cmd['-B'] = branch
        else:
            cmd['-B'] = None
    elif opts.get('move'):
        if len(args) > 0:
            if len(args) > 1:
                old = args.pop(0)
            else:
                # shell command to output the active bookmark for the active
                # revision
                old = '`hg log -T"{activebookmark}" -r .`'
        new = args[0]
        cmd['-m'] = old
        cmd.append(new)
    else:
        if len(args) > 1:
            cmd['-r'] = args[1]
            cmd.append(args[0])
        elif len(args) == 1:
            cmd.append(args[0])
    ui.status((str(cmd)), "\n")

def ispath(repo, string):
    """
    The first argument to git checkout can either be a revision or a path. Let's
    generally assume it's a revision, unless it's obviously a path. There are
    too many ways to spell revisions in git for us to reasonably catch all of
    them, so let's be conservative.
    """
    if string in repo:
        # if it's definitely a revision let's not even check if a file of the
        # same name exists.
        return False

    cwd = repo.getcwd()
    if cwd == '':
        repopath = string
    else:
        repopath = cwd + '/' + string

    exists = repo.wvfs.exists(repopath)
    if exists:
        return True

    manifest = repo['.'].manifest()

    didexist = (repopath in manifest) or manifest.hasdir(repopath)

    return didexist

def checkout(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('b', 'branch', '', ''),
        ('B', 'branch', '', ''),
        ('f', 'force', None, ''),
        ('p', 'patch', None, ''),
    ]
    paths = []
    if '--' in args:
        sepindex = args.index('--')
        paths.extend(args[sepindex + 1:])
        args = args[:sepindex]

    args, opts = parseoptions(ui, cmdoptions, args)

    rev = None
    if args and ispath(repo, args[0]):
        paths = args + paths
    elif args:
        rev = args[0]
        paths = args[1:] + paths

    cmd = Command('update')

    if opts.get('force'):
        if paths or rev:
            cmd['-C'] = None

    if opts.get('patch'):
        cmd = Command('revert')
        cmd['-i'] = None

    if opts.get('branch'):
        if len(args) == 0:
            cmd = Command('bookmark')
            cmd.append(opts.get('branch'))
        else:
            cmd.append(args[0])
            bookcmd = Command('bookmark')
            bookcmd.append(opts.get('branch'))
            cmd = cmd & bookcmd
    # if there is any path argument supplied, use revert instead of update
    elif len(paths) > 0:
        ui.status(_("note: use --no-backup to avoid creating .orig files\n\n"))
        cmd = Command('revert')
        if opts.get('patch'):
            cmd['-i'] = None
        if rev:
            cmd['-r'] = rev
        cmd.extend(paths)
    elif rev:
        if opts.get('patch'):
            cmd['-r'] = rev
        else:
            cmd.append(rev)
    elif opts.get('force'):
        cmd = Command('revert')
        cmd['--all'] = None
    else:
        raise error.Abort("a commit must be specified")

    ui.status((str(cmd)), "\n")

def cherrypick(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'continue', None, ''),
        ('', 'abort', None, ''),
        ('e', 'edit', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('graft')

    if opts.get('edit'):
        cmd['--edit'] = None
    if opts.get('continue'):
        cmd['--continue'] = None
    elif opts.get('abort'):
        ui.status(_("note: hg graft does not have --abort.\n\n"))
        return
    else:
        cmd.extend(args)

    ui.status((str(cmd)), "\n")

def clean(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('d', 'd', None, ''),
        ('f', 'force', None, ''),
        ('x', 'x', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('purge')
    if opts.get('x'):
        cmd['--all'] = None
    cmd.extend(args)

    ui.status((str(cmd)), "\n")

def clone(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'bare', None, ''),
        ('n', 'no-checkout', None, ''),
        ('b', 'branch', '', ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    if len(args) == 0:
        raise error.Abort("a repository to clone must be specified")

    cmd = Command('clone')
    cmd.append(args[0])
    if len(args) > 1:
        cmd.append(args[1])

    if opts.get('bare'):
        cmd['-U'] = None
        ui.status(_("note: Mercurial does not have bare clones. " +
            "-U will clone the repo without checking out a commit\n\n"))
    elif opts.get('no_checkout'):
        cmd['-U'] = None

    if opts.get('branch'):
        cocmd = Command("update")
        cocmd.append(opts.get('branch'))
        cmd = cmd & cocmd

    ui.status((str(cmd)), "\n")

def commit(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('a', 'all', None, ''),
        ('m', 'message', '', ''),
        ('p', 'patch', None, ''),
        ('C', 'reuse-message', '', ''),
        ('F', 'file', '', ''),
        ('', 'author', '', ''),
        ('', 'date', '', ''),
        ('', 'amend', None, ''),
        ('', 'no-edit', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('commit')
    if opts.get('patch'):
        cmd = Command('commit --interactive')

    if opts.get('amend'):
        if opts.get('no_edit'):
            cmd = Command('amend')
        else:
            cmd['--amend'] = None

    if opts.get('reuse_message'):
        cmd['-M'] = opts.get('reuse_message')

    if opts.get('message'):
        cmd['-m'] = "'%s'" % (opts.get('message'),)

    if opts.get('all'):
        ui.status(_("note: Mercurial doesn't have a staging area, " +
            "so there is no --all. -A will add and remove files " +
            "for you though.\n\n"))

    if opts.get('file'):
        cmd['-l'] = opts.get('file')

    if opts.get('author'):
        cmd['-u'] = opts.get('author')

    if opts.get('date'):
        cmd['-d'] = opts.get('date')

    cmd.extend(args)

    ui.status((str(cmd)), "\n")

def deprecated(ui, repo, *args, **kwargs):
    ui.warn(_('This command has been deprecated in the git project, ' +
        'thus isn\'t supported by this tool.\n\n'))

def diff(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('a', 'all', None, ''),
        ('', 'cached', None, ''),
        ('R', 'reverse', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('diff')

    if opts.get('cached'):
        ui.status(_('note: Mercurial has no concept of a staging area, ' +
            'so --cached does nothing.\n\n'))

    if opts.get('reverse'):
        cmd['--reverse'] = None

    for a in list(args):
        args.remove(a)
        try:
            repo.revs(a)
            cmd['-r'] = a
        except Exception:
            cmd.append(a)

    ui.status((str(cmd)), "\n")

def difftool(ui, repo, *args, **kwargs):
    ui.status(_('Mercurial does not enable external difftool by default. You '
        'need to enable the extdiff extension in your .hgrc file by adding\n'
        'extdiff =\n'
        'to the [extensions] section and then running\n\n'
        'hg extdiff -p <program>\n\n'
        'See \'hg help extdiff\' and \'hg help -e extdiff\' for more '
        'information.\n'))

def fetch(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'all', None, ''),
        ('f', 'force', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('pull')

    if len(args) > 0:
        cmd.append(args[0])
        if len(args) > 1:
            ui.status(_("note: Mercurial doesn't have refspecs. " +
                "-r can be used to specify which commits you want to pull. " +
                "-B can be used to specify which bookmark you want to pull." +
                "\n\n"))
            for v in args[1:]:
                if v in repo._bookmarks:
                    cmd['-B'] = v
                else:
                    cmd['-r'] = v

    ui.status((str(cmd)), "\n")

def grep(ui, repo, *args, **kwargs):
    cmdoptions = [
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('grep')

    # For basic usage, git grep and hg grep are the same. They both have the
    # pattern first, followed by paths.
    cmd.extend(args)

    ui.status((str(cmd)), "\n")

def init(ui, repo, *args, **kwargs):
    cmdoptions = [
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('init')

    if len(args) > 0:
        cmd.append(args[0])

    ui.status((str(cmd)), "\n")

def log(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'follow', None, ''),
        ('', 'decorate', None, ''),
        ('n', 'number', '', ''),
        ('1', '1', None, ''),
        ('', 'pretty', '', ''),
        ('', 'format', '', ''),
        ('', 'oneline', None, ''),
        ('', 'stat', None, ''),
        ('', 'graph', None, ''),
        ('p', 'patch', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)
    ui.status(_('note: -v prints the entire commit message like Git does. To ' +
              'print just the first line, drop the -v.\n\n'))
    ui.status(_("note: see hg help revset for information on how to filter " +
        "log output.\n\n"))

    cmd = Command('log')
    cmd['-v'] = None

    if opts.get('number'):
        cmd['-l'] = opts.get('number')
    if opts.get('1'):
        cmd['-l'] = '1'
    if opts.get('stat'):
        cmd['--stat'] = None
    if opts.get('graph'):
        cmd['-G'] = None
    if opts.get('patch'):
        cmd['-p'] = None

    if opts.get('pretty') or opts.get('format') or opts.get('oneline'):
        format = opts.get('format', '')
        if 'format:' in format:
            ui.status(_("note: --format format:??? equates to Mercurial's " +
                "--template. See hg help templates for more info.\n\n"))
            cmd['--template'] = '???'
        else:
            ui.status(_("note: --pretty/format/oneline equate to Mercurial's " +
                "--style or --template. See hg help templates for more info." +
                "\n\n"))
            cmd['--style'] = '???'

    if len(args) > 0:
        if '..' in args[0]:
            since, until = args[0].split('..')
            cmd['-r'] = "'%s::%s'" % (since, until)
            del args[0]
        cmd.extend(args)

    ui.status((str(cmd)), "\n")

def lsfiles(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('c', 'cached', None, ''),
        ('d', 'deleted', None, ''),
        ('m', 'modified', None, ''),
        ('o', 'others', None, ''),
        ('i', 'ignored', None, ''),
        ('s', 'stage', None, ''),
        ('z', '_zero', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    if (opts.get('modified') or opts.get('deleted')
        or opts.get('others') or opts.get('ignored')):
        cmd = Command('status')
        if opts.get('deleted'):
            cmd['-d'] = None
        if opts.get('modified'):
            cmd['-m'] = None
        if opts.get('others'):
            cmd['-o'] = None
        if opts.get('ignored'):
            cmd['-i'] = None
    else:
        cmd = Command('files')
    if opts.get('stage'):
        ui.status(_("note: Mercurial doesn't have a staging area, ignoring "
                  "--stage\n"))
    if opts.get('_zero'):
        cmd['-0'] = None
    cmd.append('.')
    for include in args:
        cmd['-I'] = util.shellquote(include)

    ui.status((str(cmd)), "\n")

def merge(ui, repo, *args, **kwargs):
    cmdoptions = [
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('merge')

    if len(args) > 0:
        cmd.append(args[len(args) - 1])

    ui.status((str(cmd)), "\n")

def mergebase(ui, repo, *args, **kwargs):
    cmdoptions = []
    args, opts = parseoptions(ui, cmdoptions, args)

    if len(args) != 2:
        args = ['A', 'B']

    cmd = Command("log -T '{node}\\n' -r 'ancestor(%s,%s)'"
                  % (args[0], args[1]))

    ui.status(_('NOTE: ancestors() is part of the revset language.\n'),
              _("Learn more about revsets with 'hg help revsets'\n\n"))
    ui.status((str(cmd)), "\n")

def mergetool(ui, repo, *args, **kwargs):
    cmdoptions = []
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command("resolve")

    if len(args) == 0:
        cmd['--all'] = None
    cmd.extend(args)
    ui.status((str(cmd)), "\n")

def mv(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('f', 'force', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('mv')
    cmd.extend(args)

    if opts.get('force'):
        cmd['-f'] = None

    ui.status((str(cmd)), "\n")

def pull(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'all', None, ''),
        ('f', 'force', None, ''),
        ('r', 'rebase', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('pull')
    cmd['--rebase'] = None

    if len(args) > 0:
        cmd.append(args[0])
        if len(args) > 1:
            ui.status(_("note: Mercurial doesn't have refspecs. " +
                "-r can be used to specify which commits you want to pull. " +
                "-B can be used to specify which bookmark you want to pull." +
                "\n\n"))
            for v in args[1:]:
                if v in repo._bookmarks:
                    cmd['-B'] = v
                else:
                    cmd['-r'] = v

    ui.status((str(cmd)), "\n")

def push(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'all', None, ''),
        ('f', 'force', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('push')

    if len(args) > 0:
        cmd.append(args[0])
        if len(args) > 1:
            ui.status(_("note: Mercurial doesn't have refspecs. " +
                "-r can be used to specify which commits you want to push. " +
                "-B can be used to specify which bookmark you want to push." +
                "\n\n"))
            for v in args[1:]:
                if v in repo._bookmarks:
                    cmd['-B'] = v
                else:
                    cmd['-r'] = v

    if opts.get('force'):
        cmd['-f'] = None

    ui.status((str(cmd)), "\n")

def rebase(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'all', None, ''),
        ('i', 'interactive', None, ''),
        ('', 'onto', '', ''),
        ('', 'abort', None, ''),
        ('', 'continue', None, ''),
        ('', 'skip', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    if opts.get('interactive'):
        ui.status(_("note: hg histedit does not perform a rebase. " +
            "It just edits history.\n\n"))
        cmd = Command('histedit')
        if len(args) > 0:
            ui.status(_("also note: 'hg histedit' will automatically detect"
                      " your stack, so no second argument is necessary.\n\n"))
        ui.status((str(cmd)), "\n")
        return

    if opts.get('skip'):
        cmd = Command('revert --all -r .')
        ui.status((str(cmd)), "\n")

    cmd = Command('rebase')

    if opts.get('continue') or opts.get('skip'):
        cmd['--continue'] = None
    if opts.get('abort'):
        cmd['--abort'] = None

    if opts.get('onto'):
        ui.status(_("note: if you're trying to lift a commit off one branch, " +
            "try hg rebase -d <destination commit> -s <commit to be lifted>" +
            "\n\n"))
        cmd['-d'] = convert(opts.get('onto'))
        if len(args) < 2:
            raise error.Abort("Expected format: git rebase --onto X Y Z")
        cmd['-s'] = "'::%s - ::%s'" % (convert(args[1]), convert(args[0]))
    else:
        if len(args) == 1:
            cmd['-d'] = convert(args[0])
        elif len(args) == 2:
            cmd['-d'] = convert(args[0])
            cmd['-b'] = convert(args[1])

    ui.status((str(cmd)), "\n")

def reflog(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'all', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('journal')
    if opts.get('all'):
        cmd['--all'] = None
    if len(args) > 0:
        cmd.append(args[0])

    ui.status(str(cmd), "\n\n")
    ui.status(_("note: in hg commits can be deleted from repo but we always"
              " have backups.\n"))

def reset(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'soft', None, ''),
        ('', 'hard', None, ''),
        ('', 'mixed', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    commit = convert(args[0] if len(args) > 0 else '.')
    hard = opts.get('hard')

    if opts.get('mixed'):
        ui.status(_('NOTE: --mixed has no meaning since Mercurial has no '
                    'staging area\n\n'))
    if opts.get('soft'):
        ui.status(_('NOTE: --soft has no meaning since Mercurial has no '
                    'staging area\n\n'))

    cmd = Command('update')
    if hard:
        cmd.append('--clean')

    cmd.append(commit)

    ui.status((str(cmd)), "\n")

def revert(ui, repo, *args, **kwargs):
    cmdoptions = [
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    if len(args) > 1:
        ui.status(_("note: hg backout doesn't support multiple commits at " +
                    "once\n\n"))

    cmd = Command('backout')
    if args:
        cmd.append(args[0])

    ui.status((str(cmd)), "\n")

def revparse(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'show-cdup', None, ''),
        ('', 'show-toplevel', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    if opts.get('show_cdup') or opts.get('show_toplevel'):
        cmd = Command('root')
        if opts.get('show_cdup'):
            ui.status(_("note: hg root prints the root of the repository\n\n"))
        ui.status((str(cmd)), "\n")
    else:
        ui.status(_("note: see hg help revset for how to refer to commits\n"))

def rm(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('f', 'force', None, ''),
        ('n', 'dry-run', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('rm')
    cmd.extend(args)

    if opts.get('force'):
        cmd['-f'] = None
    if opts.get('dry_run'):
        cmd['-n'] = None

    ui.status((str(cmd)), "\n")

def show(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'name-status', None, ''),
        ('', 'pretty', '', ''),
        ('U', 'unified', int, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    if opts.get('name_status'):
        if opts.get('pretty') == 'format:':
            cmd = Command('status')
            cmd['--change'] = '.'
        else:
            cmd = Command('log')
            cmd.append('--style status')
            cmd.append('-r .')
    elif len(args) > 0:
        if ispath(repo, args[0]):
            cmd = Command('cat')
        else:
            cmd = Command('export')
        cmd.extend(args)
        if opts.get('unified'):
            cmd.append('--config diff.unified=%d' % (opts['unified'],))
    elif opts.get('unified'):
        cmd = Command('export')
        cmd.append('--config diff.unified=%d' % (opts['unified'],))
    else:
        cmd = Command('export')

    ui.status((str(cmd)), "\n")

def stash(ui, repo, *args, **kwargs):
    cmdoptions = [
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('shelve')
    action = args[0] if len(args) > 0 else None

    if action == 'list':
        cmd['-l'] = None
    elif action == 'drop':
        cmd['-d'] = None
        if len(args) > 1:
            cmd.append(args[1])
        else:
            cmd.append('<shelve name>')
    elif action == 'pop' or action == 'apply':
        cmd = Command('unshelve')
        if len(args) > 1:
            cmd.append(args[1])
        if action == 'apply':
            cmd['--keep'] = None
    elif (action == 'branch' or action == 'show' or action == 'clear'
        or action == 'create'):
        ui.status(_("note: Mercurial doesn't have equivalents to the " +
            "git stash branch, show, clear, or create actions.\n\n"))
        return
    else:
        if len(args) > 0:
            if args[0] != 'save':
                cmd['--name'] = args[0]
            elif len(args) > 1:
                cmd['--name'] = args[1]

    ui.status((str(cmd)), "\n")

def status(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('', 'ignored', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('status')
    cmd.extend(args)

    if opts.get('ignored'):
        cmd['-i'] = None

    ui.status((str(cmd)), "\n")

def svn(ui, repo, *args, **kwargs):
    svncmd = args[0]
    if not svncmd in gitsvncommands:
        ui.warn(_("error: unknown git svn command %s\n") % (svncmd))

    args = args[1:]
    return gitsvncommands[svncmd](ui, repo, *args, **kwargs)

def svndcommit(ui, repo, *args, **kwargs):
    cmdoptions = [
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('push')

    ui.status((str(cmd)), "\n")

def svnfetch(ui, repo, *args, **kwargs):
    cmdoptions = [
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('pull')
    cmd.append('default-push')

    ui.status((str(cmd)), "\n")

def svnfindrev(ui, repo, *args, **kwargs):
    cmdoptions = [
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    cmd = Command('log')
    cmd['-r'] = args[0]

    ui.status((str(cmd)), "\n")

def svnrebase(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('l', 'local', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    pullcmd = Command('pull')
    pullcmd.append('default-push')
    rebasecmd = Command('rebase')
    rebasecmd.append('tip')

    cmd = pullcmd & rebasecmd

    ui.status((str(cmd)), "\n")

def tag(ui, repo, *args, **kwargs):
    cmdoptions = [
        ('f', 'force', None, ''),
        ('l', 'list', None, ''),
        ('d', 'delete', None, ''),
    ]
    args, opts = parseoptions(ui, cmdoptions, args)

    if opts.get('list'):
        cmd = Command('tags')
    else:
        cmd = Command('tag')
        cmd.append(args[0])
        if len(args) > 1:
            cmd['-r'] = args[1]

        if opts.get('delete'):
            cmd['--remove'] = None

        if opts.get('force'):
            cmd['-f'] = None

    ui.status((str(cmd)), "\n")

gitcommands = {
    'add': add,
    'am': am,
    'apply': apply,
    'bisect': bisect,
    'blame': blame,
    'branch': branch,
    'checkout': checkout,
    'cherry-pick': cherrypick,
    'clean': clean,
    'clone': clone,
    'commit': commit,
    'diff': diff,
    'difftool': difftool,
    'fetch': fetch,
    'grep': grep,
    'init': init,
    'log': log,
    'ls-files': lsfiles,
    'merge': merge,
    'merge-base': mergebase,
    'mergetool': mergetool,
    'mv': mv,
    'pull': pull,
    'push': push,
    'rebase': rebase,
    'reflog': reflog,
    'reset': reset,
    'revert': revert,
    'rev-parse': revparse,
    'rm': rm,
    'show': show,
    'stash': stash,
    'status': status,
    'svn': svn,
    'tag': tag,
    'whatchanged': deprecated,
}

gitsvncommands = {
    'dcommit': svndcommit,
    'fetch': svnfetch,
    'find-rev': svnfindrev,
    'rebase': svnrebase,
}



# bdiff.py - Python implementation of bdiff.c
#
# Copyright 2009 Matt Mackall <mpm@selenic.com> and others
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import difflib
import re
import struct

def splitnewlines(text):
    '''like str.splitlines, but only split on newlines.'''
    lines = [l + '\n' for l in text.split('\n')]
    if lines:
        if lines[-1] == '\n':
            lines.pop()
        else:
            lines[-1] = lines[-1][:-1]
    return lines

def _normalizeblocks(a, b, blocks):
    prev = None
    r = []
    for curr in blocks:
        if prev is None:
            prev = curr
            continue
        shift = 0

        a1, b1, l1 = prev
        a1end = a1 + l1
        b1end = b1 + l1

        a2, b2, l2 = curr
        a2end = a2 + l2
        b2end = b2 + l2
        if a1end == a2:
            while (a1end + shift < a2end and
                   a[a1end + shift] == b[b1end + shift]):
                shift += 1
        elif b1end == b2:
            while (b1end + shift < b2end and
                   a[a1end + shift] == b[b1end + shift]):
                shift += 1
        r.append((a1, b1, l1 + shift))
        prev = a2 + shift, b2 + shift, l2 - shift
    r.append(prev)
    return r

def bdiff(a, b):
    a = bytes(a).splitlines(True)
    b = bytes(b).splitlines(True)

    if not a:
        s = "".join(b)
        return s and (struct.pack(">lll", 0, 0, len(s)) + s)

    bin = []
    p = [0]
    for i in a:
        p.append(p[-1] + len(i))

    d = difflib.SequenceMatcher(None, a, b).get_matching_blocks()
    d = _normalizeblocks(a, b, d)
    la = 0
    lb = 0
    for am, bm, size in d:
        s = "".join(b[lb:bm])
        if am > la or s:
            bin.append(struct.pack(">lll", p[la], p[am], len(s)) + s)
        la = am + size
        lb = bm + size

    return "".join(bin)

def blocks(a, b):
    an = splitnewlines(a)
    bn = splitnewlines(b)
    d = difflib.SequenceMatcher(None, an, bn).get_matching_blocks()
    d = _normalizeblocks(an, bn, d)
    return [(i, i + n, j, j + n) for (i, j, n) in d]

def fixws(text, allws):
    if allws:
        text = re.sub('[ \t\r]+', '', text)
    else:
        text = re.sub('[ \t\r]+', ' ', text)
        text = text.replace(' \n', '\n')
    return text



# hgweb/__init__.py - web interface to a mercurial repository
#
# Copyright 21 May 2005 - (c) 2005 Jake Edge <jake@edge2.net>
# Copyright 2005 Matt Mackall <mpm@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import os

from ..i18n import _

from .. import (
    error,
    pycompat,
    util,
)

from . import (
    hgweb_mod,
    hgwebdir_mod,
    server,
)

def hgweb(config, name=None, baseui=None):
    '''create an hgweb wsgi object

    config can be one of:
    - repo object (single repo view)
    - path to repo (single repo view)
    - path to config file (multi-repo view)
    - dict of virtual:real pairs (multi-repo view)
    - list of virtual:real tuples (multi-repo view)
    '''

    if ((isinstance(config, str) and not os.path.isdir(config)) or
        isinstance(config, dict) or isinstance(config, list)):
        # create a multi-dir interface
        return hgwebdir_mod.hgwebdir(config, baseui=baseui)
    return hgweb_mod.hgweb(config, name=name, baseui=baseui)

def hgwebdir(config, baseui=None):
    return hgwebdir_mod.hgwebdir(config, baseui=baseui)

class httpservice(object):
    def __init__(self, ui, app, opts):
        self.ui = ui
        self.app = app
        self.opts = opts

    def init(self):
        util.setsignalhandler()
        self.httpd = server.create_server(self.ui, self.app)

        if self.opts['port'] and not self.ui.verbose:
            return

        if self.httpd.prefix:
            prefix = self.httpd.prefix.strip('/') + '/'
        else:
            prefix = ''

        port = r':%d' % self.httpd.port
        if port == r':80':
            port = r''

        bindaddr = self.httpd.addr
        if bindaddr == r'0.0.0.0':
            bindaddr = r'*'
        elif r':' in bindaddr: # IPv6
            bindaddr = r'[%s]' % bindaddr

        fqaddr = self.httpd.fqaddr
        if r':' in fqaddr:
            fqaddr = r'[%s]' % fqaddr
        if self.opts['port']:
            write = self.ui.status
        else:
            write = self.ui.write
        write(_('listening at http://%s%s/%s (bound to %s:%d)\n') %
              (pycompat.sysbytes(fqaddr), pycompat.sysbytes(port),
               prefix, pycompat.sysbytes(bindaddr), self.httpd.port))
        self.ui.flush()  # avoid buffering of status message

    def run(self):
        self.httpd.serve_forever()

def createapp(baseui, repo, webconf):
    if webconf:
        return hgwebdir_mod.hgwebdir(webconf, baseui=baseui)
    else:
        if not repo:
            raise error.RepoError(_("there is no Mercurial repository"
                                    " here (.hg not found)"))
        return hgweb_mod.hgweb(repo, baseui=baseui)



# parsers.py - Python implementation of parsers.c
#
# Copyright 2009 Matt Mackall <mpm@selenic.com> and others
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import struct
import zlib

from ..node import nullid
from .. import pycompat
stringio = pycompat.stringio


_pack = struct.pack
_unpack = struct.unpack
_compress = zlib.compress
_decompress = zlib.decompress

# Some code below makes tuples directly because it's more convenient. However,
# code outside this module should always use dirstatetuple.
def dirstatetuple(*x):
    # x is a tuple
    return x

indexformatng = ">Qiiiiii20s12x"
indexfirst = struct.calcsize('Q')
sizeint = struct.calcsize('i')
indexsize = struct.calcsize(indexformatng)

def gettype(q):
    return int(q & 0xFFFF)

def offset_type(offset, type):
    return int(int(offset) << 16 | type)

class BaseIndexObject(object):
    def __len__(self):
        return self._lgt + len(self._extra) + 1

    def insert(self, i, tup):
        assert i == -1
        self._extra.append(tup)

    def _fix_index(self, i):
        if not isinstance(i, int):
            raise TypeError("expecting int indexes")
        if i < 0:
            i = len(self) + i
        if i < 0 or i >= len(self):
            raise IndexError
        return i

    def __getitem__(self, i):
        i = self._fix_index(i)
        if i == len(self) - 1:
            return (0, 0, 0, -1, -1, -1, -1, nullid)
        if i >= self._lgt:
            return self._extra[i - self._lgt]
        index = self._calculate_index(i)
        r = struct.unpack(indexformatng, self._data[index:index + indexsize])
        if i == 0:
            e = list(r)
            type = gettype(e[0])
            e[0] = offset_type(0, type)
            return tuple(e)
        return r

class IndexObject(BaseIndexObject):
    def __init__(self, data):
        assert len(data) % indexsize == 0
        self._data = data
        self._lgt = len(data) // indexsize
        self._extra = []

    def _calculate_index(self, i):
        return i * indexsize

    def __delitem__(self, i):
        if not isinstance(i, slice) or not i.stop == -1 or i.step is not None:
            raise ValueError("deleting slices only supports a:-1 with step 1")
        i = self._fix_index(i.start)
        if i < self._lgt:
            self._data = self._data[:i * indexsize]
            self._lgt = i
            self._extra = []
        else:
            self._extra = self._extra[:i - self._lgt]

class InlinedIndexObject(BaseIndexObject):
    def __init__(self, data, inline=0):
        self._data = data
        self._lgt = self._inline_scan(None)
        self._inline_scan(self._lgt)
        self._extra = []

    def _inline_scan(self, lgt):
        off = 0
        if lgt is not None:
            self._offsets = [0] * lgt
        count = 0
        while off <= len(self._data) - indexsize:
            s, = struct.unpack('>i',
                self._data[off + indexfirst:off + sizeint + indexfirst])
            if lgt is not None:
                self._offsets[count] = off
            count += 1
            off += indexsize + s
        if off != len(self._data):
            raise ValueError("corrupted data")
        return count

    def __delitem__(self, i):
        if not isinstance(i, slice) or not i.stop == -1 or i.step is not None:
            raise ValueError("deleting slices only supports a:-1 with step 1")
        i = self._fix_index(i.start)
        if i < self._lgt:
            self._offsets = self._offsets[:i]
            self._lgt = i
            self._extra = []
        else:
            self._extra = self._extra[:i - self._lgt]

    def _calculate_index(self, i):
        return self._offsets[i]

def parse_index2(data, inline):
    if not inline:
        return IndexObject(data), None
    return InlinedIndexObject(data, inline), (0, data)

def parse_dirstate(dmap, copymap, st):
    parents = [st[:20], st[20: 40]]
    # dereference fields so they will be local in loop
    format = ">cllll"
    e_size = struct.calcsize(format)
    pos1 = 40
    l = len(st)

    # the inner loop
    while pos1 < l:
        pos2 = pos1 + e_size
        e = _unpack(">cllll", st[pos1:pos2]) # a literal here is faster
        pos1 = pos2 + e[4]
        f = st[pos2:pos1]
        if '\0' in f:
            f, c = f.split('\0')
            copymap[f] = c
        dmap[f] = e[:4]
    return parents

def pack_dirstate(dmap, copymap, pl, now):
    now = int(now)
    cs = stringio()
    write = cs.write
    write("".join(pl))
    for f, e in dmap.iteritems():
        if e[0] == 'n' and e[3] == now:
            # The file was last modified "simultaneously" with the current
            # write to dirstate (i.e. within the same second for file-
            # systems with a granularity of 1 sec). This commonly happens
            # for at least a couple of files on 'update'.
            # The user could change the file without changing its size
            # within the same second. Invalidate the file's mtime in
            # dirstate, forcing future 'status' calls to compare the
            # contents of the file if the size is the same. This prevents
            # mistakenly treating such files as clean.
            e = dirstatetuple(e[0], e[1], e[2], -1)
            dmap[f] = e

        if f in copymap:
            f = "%s\0%s" % (f, copymap[f])
        e = _pack(">cllll", e[0], e[1], e[2], e[3], len(f))
        write(e)
        write(f)
    return cs.getvalue()



# repository.py - Interfaces and base classes for repositories and peers.
#
# Copyright 2017 Gregory Szorc <gregory.szorc@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import abc

from .i18n import _
from . import (
    error,
)

class _basepeer(object):
    """Represents a "connection" to a repository.

    This is the base interface for representing a connection to a repository.
    It holds basic properties and methods applicable to all peer types.

    This is not a complete interface definition and should not be used
    outside of this module.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def ui(self):
        """ui.ui instance."""

    @abc.abstractmethod
    def url(self):
        """Returns a URL string representing this peer.

        Currently, implementations expose the raw URL used to construct the
        instance. It may contain credentials as part of the URL. The
        expectations of the value aren't well-defined and this could lead to
        data leakage.

        TODO audit/clean consumers and more clearly define the contents of this
        value.
        """

    @abc.abstractmethod
    def local(self):
        """Returns a local repository instance.

        If the peer represents a local repository, returns an object that
        can be used to interface with it. Otherwise returns ``None``.
        """

    @abc.abstractmethod
    def peer(self):
        """Returns an object conforming to this interface.

        Most implementations will ``return self``.
        """

    @abc.abstractmethod
    def canpush(self):
        """Returns a boolean indicating if this peer can be pushed to."""

    @abc.abstractmethod
    def close(self):
        """Close the connection to this peer.

        This is called when the peer will no longer be used. Resources
        associated with the peer should be cleaned up.
        """

class _basewirecommands(object):
    """Client-side interface for communicating over the wire protocol.

    This interface is used as a gateway to the Mercurial wire protocol.
    methods commonly call wire protocol commands of the same name.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def branchmap(self):
        """Obtain heads in named branches.

        Returns a dict mapping branch name to an iterable of nodes that are
        heads on that branch.
        """

    @abc.abstractmethod
    def capabilities(self):
        """Obtain capabilities of the peer.

        Returns a set of string capabilities.
        """

    @abc.abstractmethod
    def debugwireargs(self, one, two, three=None, four=None, five=None):
        """Used to facilitate debugging of arguments passed over the wire."""

    @abc.abstractmethod
    def getbundle(self, source, **kwargs):
        """Obtain remote repository data as a bundle.

        This command is how the bulk of repository data is transferred from
        the peer to the local repository

        Returns a generator of bundle data.
        """

    @abc.abstractmethod
    def heads(self):
        """Determine all known head revisions in the peer.

        Returns an iterable of binary nodes.
        """

    @abc.abstractmethod
    def known(self, nodes):
        """Determine whether multiple nodes are known.

        Accepts an iterable of nodes whose presence to check for.

        Returns an iterable of booleans indicating of the corresponding node
        at that index is known to the peer.
        """

    @abc.abstractmethod
    def listkeys(self, namespace):
        """Obtain all keys in a pushkey namespace.

        Returns an iterable of key names.
        """

    @abc.abstractmethod
    def lookup(self, key):
        """Resolve a value to a known revision.

        Returns a binary node of the resolved revision on success.
        """

    @abc.abstractmethod
    def pushkey(self, namespace, key, old, new):
        """Set a value using the ``pushkey`` protocol.

        Arguments correspond to the pushkey namespace and key to operate on and
        the old and new values for that key.

        Returns a string with the peer result. The value inside varies by the
        namespace.
        """

    @abc.abstractmethod
    def stream_out(self):
        """Obtain streaming clone data.

        Successful result should be a generator of data chunks.
        """

    @abc.abstractmethod
    def unbundle(self, bundle, heads, url):
        """Transfer repository data to the peer.

        This is how the bulk of data during a push is transferred.

        Returns the integer number of heads added to the peer.
        """

class _baselegacywirecommands(object):
    """Interface for implementing support for legacy wire protocol commands.

    Wire protocol commands transition to legacy status when they are no longer
    used by modern clients. To facilitate identifying which commands are
    legacy, the interfaces are split.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def between(self, pairs):
        """Obtain nodes between pairs of nodes.

        ``pairs`` is an iterable of node pairs.

        Returns an iterable of iterables of nodes corresponding to each
        requested pair.
        """

    @abc.abstractmethod
    def branches(self, nodes):
        """Obtain ancestor changesets of specific nodes back to a branch point.

        For each requested node, the peer finds the first ancestor node that is
        a DAG root or is a merge.

        Returns an iterable of iterables with the resolved values for each node.
        """

    @abc.abstractmethod
    def changegroup(self, nodes, kind):
        """Obtain a changegroup with data for descendants of specified nodes."""

    @abc.abstractmethod
    def changegroupsubset(self, bases, heads, kind):
        pass

class peer(_basepeer, _basewirecommands):
    """Unified interface and base class for peer repositories.

    All peer instances must inherit from this class and conform to its
    interface.
    """

    @abc.abstractmethod
    def iterbatch(self):
        """Obtain an object to be used for multiple method calls.

        Various operations call several methods on peer instances. If each
        method call were performed immediately and serially, this would
        require round trips to remote peers and/or would slow down execution.

        Some peers have the ability to "batch" method calls to avoid costly
        round trips or to facilitate concurrent execution.

        This method returns an object that can be used to indicate intent to
        perform batched method calls.

        The returned object is a proxy of this peer. It intercepts calls to
        batchable methods and queues them instead of performing them
        immediately. This proxy object has a ``submit`` method that will
        perform all queued batchable method calls. A ``results()`` method
        exposes the results of queued/batched method calls. It is a generator
        of results in the order they were called.

        Not all peers or wire protocol implementations may actually batch method
        calls. However, they must all support this API.
        """

    def capable(self, name):
        """Determine support for a named capability.

        Returns ``False`` if capability not supported.

        Returns ``True`` if boolean capability is supported. Returns a string
        if capability support is non-boolean.
        """
        caps = self.capabilities()
        if name in caps:
            return True

        name = '%s=' % name
        for cap in caps:
            if cap.startswith(name):
                return cap[len(name):]

        return False

    def requirecap(self, name, purpose):
        """Require a capability to be present.

        Raises a ``CapabilityError`` if the capability isn't present.
        """
        if self.capable(name):
            return

        raise error.CapabilityError(
            _('cannot %s; remote repository does not support the %r '
              'capability') % (purpose, name))

class legacypeer(peer, _baselegacywirecommands):
    """peer but with support for legacy wire protocol commands."""



# peer.py - repository base classes for mercurial
#
# Copyright 2005, 2006 Matt Mackall <mpm@selenic.com>
# Copyright 2006 Vadim Gelfer <vadim.gelfer@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

from . import (
    error,
    pycompat,
    util,
)

# abstract batching support

class future(object):
    '''placeholder for a value to be set later'''
    def set(self, value):
        if util.safehasattr(self, 'value'):
            raise error.RepoError("future is already set")
        self.value = value

class batcher(object):
    '''base class for batches of commands submittable in a single request

    All methods invoked on instances of this class are simply queued and
    return a a future for the result. Once you call submit(), all the queued
    calls are performed and the results set in their respective futures.
    '''
    def __init__(self):
        self.calls = []
    def __getattr__(self, name):
        def call(*args, **opts):
            resref = future()
            # Please don't invent non-ascii method names, or you will
            # give core hg a very sad time.
            self.calls.append((name.encode('ascii'), args, opts, resref,))
            return resref
        return call
    def submit(self):
        raise NotImplementedError()

class iterbatcher(batcher):

    def submit(self):
        raise NotImplementedError()

    def results(self):
        raise NotImplementedError()

class localiterbatcher(iterbatcher):
    def __init__(self, local):
        super(iterbatcher, self).__init__()
        self.local = local

    def submit(self):
        # submit for a local iter batcher is a noop
        pass

    def results(self):
        for name, args, opts, resref in self.calls:
            resref.set(getattr(self.local, name)(*args, **opts))
            yield resref.value

def batchable(f):
    '''annotation for batchable methods

    Such methods must implement a coroutine as follows:

    @batchable
    def sample(self, one, two=None):
        # Build list of encoded arguments suitable for your wire protocol:
        encargs = [('one', encode(one),), ('two', encode(two),)]
        # Create future for injection of encoded result:
        encresref = future()
        # Return encoded arguments and future:
        yield encargs, encresref
        # Assuming the future to be filled with the result from the batched
        # request now. Decode it:
        yield decode(encresref.value)

    The decorator returns a function which wraps this coroutine as a plain
    method, but adds the original method as an attribute called "batchable",
    which is used by remotebatch to split the call into separate encoding and
    decoding phases.
    '''
    def plain(*args, **opts):
        batchable = f(*args, **opts)
        encargsorres, encresref = next(batchable)
        if not encresref:
            return encargsorres # a local result in this case
        self = args[0]
        cmd = pycompat.bytesurl(f.__name__)  # ensure cmd is ascii bytestr
        encresref.set(self._submitone(cmd, encargsorres))
        return next(batchable)
    setattr(plain, 'batchable', f)
    return plain



# highlight.py - highlight extension implementation file
#
#  Copyright 2007-2009 Adam Hupp <adam@hupp.org> and others
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# The original module was split in an interface and an implementation
# file to defer pygments loading and speedup extension setup.

from __future__ import absolute_import

from mercurial import demandimport
demandimport.ignore.extend(['pkgutil', 'pkg_resources', '__main__'])

from mercurial import (
    encoding,
    util,
)

with demandimport.deactivated():
    import pygments
    import pygments.formatters
    import pygments.lexers
    import pygments.plugin
    import pygments.util

    for unused in pygments.plugin.find_plugin_lexers():
        pass

highlight = pygments.highlight
ClassNotFound = pygments.util.ClassNotFound
guess_lexer = pygments.lexers.guess_lexer
guess_lexer_for_filename = pygments.lexers.guess_lexer_for_filename
TextLexer = pygments.lexers.TextLexer
HtmlFormatter = pygments.formatters.HtmlFormatter

SYNTAX_CSS = ('\n<link rel="stylesheet" href="{url}highlightcss" '
              'type="text/css" />')

def pygmentize(field, fctx, style, tmpl, guessfilenameonly=False):

    # append a <link ...> to the syntax highlighting css
    old_header = tmpl.load('header')
    if SYNTAX_CSS not in old_header:
        new_header = old_header + SYNTAX_CSS
        tmpl.cache['header'] = new_header

    text = fctx.data()
    if util.binary(text):
        return

    # str.splitlines() != unicode.splitlines() because "reasons"
    for c in "\x0c\x1c\x1d\x1e":
        if c in text:
            text = text.replace(c, '')

    # Pygments is best used with Unicode strings:
    # <http://pygments.org/docs/unicode/>
    text = text.decode(encoding.encoding, 'replace')

    # To get multi-line strings right, we can't format line-by-line
    try:
        lexer = guess_lexer_for_filename(fctx.path(), text[:1024],
                                         stripnl=False)
    except (ClassNotFound, ValueError):
        # guess_lexer will return a lexer if *any* lexer matches. There is
        # no way to specify a minimum match score. This can give a high rate of
        # false positives on files with an unknown filename pattern.
        if guessfilenameonly:
            return

        try:
            lexer = guess_lexer(text[:1024], stripnl=False)
        except (ClassNotFound, ValueError):
            # Don't highlight unknown files
            return

    # Don't highlight text files
    if isinstance(lexer, TextLexer):
        return

    formatter = HtmlFormatter(nowrap=True, style=style)

    colorized = highlight(text, lexer, formatter)
    coloriter = (s.encode(encoding.encoding, 'replace')
                 for s in colorized.splitlines())

    tmpl.filters['colorize'] = lambda x: next(coloriter)

    oldl = tmpl.cache[field]
    newl = oldl.replace('line|escape', 'line|colorize')
    tmpl.cache[field] = newl



# Copyright 2009-2010 Gregory P. Ward
# Copyright 2009-2010 Intelerad Medical Systems Incorporated
# Copyright 2010-2011 Fog Creek Software
# Copyright 2010-2011 Unity Technologies
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

'''setup for largefiles extension: uisetup'''
from __future__ import absolute_import

from mercurial.i18n import _

from mercurial.hgweb import (
    webcommands,
)

from mercurial import (
    archival,
    cmdutil,
    commands,
    copies,
    exchange,
    extensions,
    filemerge,
    hg,
    httppeer,
    merge,
    scmutil,
    sshpeer,
    subrepo,
    upgrade,
    url,
    wireproto,
)

from . import (
    overrides,
    proto,
)

def uisetup(ui):
    # Disable auto-status for some commands which assume that all
    # files in the result are under Mercurial's control

    entry = extensions.wrapcommand(commands.table, 'add',
                                   overrides.overrideadd)
    addopt = [('', 'large', None, _('add as largefile')),
              ('', 'normal', None, _('add as normal file')),
              ('', 'lfsize', '', _('add all files above this size '
                                   '(in megabytes) as largefiles '
                                   '(default: 10)'))]
    entry[1].extend(addopt)

    # The scmutil function is called both by the (trivial) addremove command,
    # and in the process of handling commit -A (issue3542)
    extensions.wrapfunction(scmutil, 'addremove', overrides.scmutiladdremove)
    extensions.wrapfunction(cmdutil, 'add', overrides.cmdutiladd)
    extensions.wrapfunction(cmdutil, 'remove', overrides.cmdutilremove)
    extensions.wrapfunction(cmdutil, 'forget', overrides.cmdutilforget)

    extensions.wrapfunction(copies, 'pathcopies', overrides.copiespathcopies)

    extensions.wrapfunction(upgrade, 'preservedrequirements',
                            overrides.upgraderequirements)

    extensions.wrapfunction(upgrade, 'supporteddestrequirements',
                            overrides.upgraderequirements)

    # Subrepos call status function
    entry = extensions.wrapcommand(commands.table, 'status',
                                   overrides.overridestatus)
    extensions.wrapfunction(subrepo.hgsubrepo, 'status',
                            overrides.overridestatusfn)

    entry = extensions.wrapcommand(commands.table, 'log',
                                   overrides.overridelog)
    entry = extensions.wrapcommand(commands.table, 'rollback',
                                   overrides.overriderollback)
    entry = extensions.wrapcommand(commands.table, 'verify',
                                   overrides.overrideverify)

    verifyopt = [('', 'large', None,
                  _('verify that all largefiles in current revision exists')),
                 ('', 'lfa', None,
                  _('verify largefiles in all revisions, not just current')),
                 ('', 'lfc', None,
                  _('verify local largefile contents, not just existence'))]
    entry[1].extend(verifyopt)

    entry = extensions.wrapcommand(commands.table, 'debugstate',
                                   overrides.overridedebugstate)
    debugstateopt = [('', 'large', None, _('display largefiles dirstate'))]
    entry[1].extend(debugstateopt)

    outgoing = lambda orgfunc, *arg, **kwargs: orgfunc(*arg, **kwargs)
    entry = extensions.wrapcommand(commands.table, 'outgoing', outgoing)
    outgoingopt = [('', 'large', None, _('display outgoing largefiles'))]
    entry[1].extend(outgoingopt)
    cmdutil.outgoinghooks.add('largefiles', overrides.outgoinghook)
    entry = extensions.wrapcommand(commands.table, 'summary',
                                   overrides.overridesummary)
    summaryopt = [('', 'large', None, _('display outgoing largefiles'))]
    entry[1].extend(summaryopt)
    cmdutil.summaryremotehooks.add('largefiles', overrides.summaryremotehook)

    entry = extensions.wrapcommand(commands.table, 'pull',
                                   overrides.overridepull)
    pullopt = [('', 'all-largefiles', None,
                 _('download all pulled versions of largefiles (DEPRECATED)')),
               ('', 'lfrev', [],
                _('download largefiles for these revisions'), _('REV'))]
    entry[1].extend(pullopt)

    entry = extensions.wrapcommand(commands.table, 'push',
                                   overrides.overridepush)
    pushopt = [('', 'lfrev', [],
                _('upload largefiles for these revisions'), _('REV'))]
    entry[1].extend(pushopt)
    extensions.wrapfunction(exchange, 'pushoperation',
                            overrides.exchangepushoperation)

    entry = extensions.wrapcommand(commands.table, 'clone',
                                   overrides.overrideclone)
    cloneopt = [('', 'all-largefiles', None,
                 _('download all versions of all largefiles'))]
    entry[1].extend(cloneopt)
    extensions.wrapfunction(hg, 'clone', overrides.hgclone)
    extensions.wrapfunction(hg, 'postshare', overrides.hgpostshare)

    entry = extensions.wrapcommand(commands.table, 'cat',
                                   overrides.overridecat)
    extensions.wrapfunction(merge, '_checkunknownfile',
                            overrides.overridecheckunknownfile)
    extensions.wrapfunction(merge, 'calculateupdates',
                            overrides.overridecalculateupdates)
    extensions.wrapfunction(merge, 'recordupdates',
                            overrides.mergerecordupdates)
    extensions.wrapfunction(merge, 'update', overrides.mergeupdate)
    extensions.wrapfunction(filemerge, '_filemerge',
                            overrides.overridefilemerge)
    extensions.wrapfunction(cmdutil, 'copy', overrides.overridecopy)

    # Summary calls dirty on the subrepos
    extensions.wrapfunction(subrepo.hgsubrepo, 'dirty', overrides.overridedirty)

    extensions.wrapfunction(cmdutil, 'revert', overrides.overriderevert)

    extensions.wrapcommand(commands.table, 'archive',
                           overrides.overridearchivecmd)
    extensions.wrapfunction(archival, 'archive', overrides.overridearchive)
    extensions.wrapfunction(subrepo.hgsubrepo, 'archive',
                            overrides.hgsubrepoarchive)
    extensions.wrapfunction(webcommands, 'archive', overrides.hgwebarchive)
    extensions.wrapfunction(cmdutil, 'bailifchanged',
                            overrides.overridebailifchanged)

    extensions.wrapfunction(cmdutil, 'postcommitstatus',
                            overrides.postcommitstatus)
    extensions.wrapfunction(scmutil, 'marktouched',
                            overrides.scmutilmarktouched)

    extensions.wrapfunction(url, 'open',
                            overrides.openlargefile)

    # create the new wireproto commands ...
    wireproto.commands['putlfile'] = (proto.putlfile, 'sha')
    wireproto.commands['getlfile'] = (proto.getlfile, 'sha')
    wireproto.commands['statlfile'] = (proto.statlfile, 'sha')

    # ... and wrap some existing ones
    wireproto.commands['heads'] = (proto.heads, '')
    wireproto.commands['lheads'] = (wireproto.heads, '')

    # make putlfile behave the same as push and {get,stat}lfile behave
    # the same as pull w.r.t. permissions checks
    wireproto.permissions['putlfile'] = 'push'
    wireproto.permissions['getlfile'] = 'pull'
    wireproto.permissions['statlfile'] = 'pull'
    wireproto.permissions['lheads'] = 'pull'

    extensions.wrapfunction(webcommands, 'decodepath', overrides.decodepath)

    extensions.wrapfunction(wireproto, '_capabilities', proto._capabilities)

    # can't do this in reposetup because it needs to have happened before
    # wirerepo.__init__ is called
    proto.ssholdcallstream = sshpeer.sshpeer._callstream
    proto.httpoldcallstream = httppeer.httppeer._callstream
    sshpeer.sshpeer._callstream = proto.sshrepocallstream
    httppeer.httppeer._callstream = proto.httprepocallstream

    # override some extensions' stuff as well
    for name, module in extensions.extensions():
        if name == 'purge':
            extensions.wrapcommand(getattr(module, 'cmdtable'), 'purge',
                overrides.overridepurge)
        if name == 'rebase':
            extensions.wrapcommand(getattr(module, 'cmdtable'), 'rebase',
                overrides.overriderebase)
            extensions.wrapfunction(module, 'rebase',
                                    overrides.overriderebase)
        if name == 'transplant':
            extensions.wrapcommand(getattr(module, 'cmdtable'), 'transplant',
                overrides.overridetransplant)
